#include <OneWire.h>
#include <DallasTemperature.h>
#include <avr/wdt.h>

// constants
const int oneWirePin = 2;
const int pwmPins[] = {5,9,10,11};
const int possibleTachoPins[] = {4,7,8,12};
const float temperaturePrecisionInDegree = 0.25; // allowed values as 0.5, 0.25, 0.125 and 0.0625
const float tempTheFansStartsToIncPWM = 30.0;   // at 30 it runs at min pwm
const float tempTheFanShouldSetToMaxPWM = 37.5; // at 39 it runs at max pwm
const int diffThreshold = 25;
const int MeasureIntervalInSeconds = 5;

// variables
int numberOfTemperatureSensors = 0;
int tachoPins[] = {-1, -1, -1, -1};
int minPWMs[] = {0, 0, 0, 0};
int pwmValues[] = {-1, -1, -1, -1};
int minRPMs[] = {0, 0, 0, 0};
int maxRPMs[] = {0, 0, 0, 0};
float lastTemp = 0;
float lastMeasure = millis();

OneWire oneWire(oneWirePin);
DallasTemperature sensors(&oneWire);

void softReset(uint8_t msToReset)
{
    wdt_enable(msToReset); // start watchdog
    while(true){};         // wait without sending anything, so wdt_reset() is called.
}

void setPWM(int pin, int val)
{
    //Serial.print("set PWM called with pin "); Serial.print(pin); Serial.print(" value "); Serial.println(val);
    val = (val<0) ? 0 : (val>100) ? 100 : val;  // only 0-100 is allowed
    int value = ((val*320)/100);  // re-calc to 0-320

    switch(pin)
    {
        case  5: OCR3A = value; break;
        case  9: OCR1A = value; break;
        case 10: OCR1B = value; break;
        case 11: OCR1C = value; break;
    }
}

float getHighestTemp()
{
    //Serial.print("Requesting temperatures...");
    sensors.requestTemperatures(); // Send the command to get temperatures
    //Serial.println("DONE");

    //Serial.println(numberOfTemperatureSensors);
    float highest = -1;

    for (int i=0 ; i<numberOfTemperatureSensors ; i++)
    {
        float tempC = sensors.getTempCByIndex(i);
        //Serial.print("temp"); Serial.print(i); Serial.print(" is "); Serial.println(tempC);
        highest = (highest<tempC) ? tempC : highest;
    }

    return (highest==-1) ?  tempTheFanShouldSetToMaxPWM : highest;
}

void setupTemperatureSensors()
{
    while(true)
    {
        sensors.begin();
        numberOfTemperatureSensors = sensors.getDS18Count();
        //Serial.print("found "); Serial.print(numberOfTemperatureSensors); Serial.println(" DS18b20 sensors.");

        if (numberOfTemperatureSensors>0) break;

        Serial.println("need at least one DS18b20 sensor. none found. check wires. retry in 5s");
        delay(5000);
    }

    if (temperaturePrecisionInDegree == 0.0625)
    {
        sensors.setResolution(0x7F);
    } 
    else if (temperaturePrecisionInDegree == 0.125)
    {
        sensors.setResolution(0x5F);
    }
    else if (temperaturePrecisionInDegree == 0.25)
    {
        sensors.setResolution(0x3F);
    }
    else
    {
        sensors.setResolution(0x1F);
    }

    Serial.print("Found "); Serial.print(numberOfTemperatureSensors); Serial.println(" x DS18b20 sensors.");
    // first values most crap, so we request two times
    sensors.requestTemperatures();
    delay(500);
    sensors.requestTemperatures();

    //lastTemp = getHighestTemp();
}

int measureRPM(int pinNumber, int roundedBy = 10)
{
    if (pinNumber<0 || pinNumber>13) return 0;
    unsigned long edgeTime = pulseIn(pinNumber, LOW);
    int rpm = ((1000/((edgeTime * 4)/1000))*6*10);
    //Serial.print("tacho on pin "); Serial.print(pinNumber); Serial.print(" has "); Serial.print(rpm); Serial.println(" RPM");
    rpm = (rpm<0) ? 0 : rpm;
    return ((rpm / roundedBy) * roundedBy);
}

bool foudAtleastOneFan()
{
    bool noneFans = ((tachoPins[0] == -1) && (tachoPins[1] == -1) && (tachoPins[2] == -1) && (tachoPins[3] == -1));
    return !noneFans;
}

void assignPinToFan()
{
    while(true)
    {
        Serial.println("checking tacho pins for signals...");

        //Serial.println("set all fans pwm to 0.");
        for (int fan=0 ; fan<4 ; fan++) setPWM(pwmPins[fan],0);
        delay(3000); // more time gives fan more time to slow down

        //Serial.println("collect min RPMs");
        // problem: measure not 100 correct. we need the highest rpm for pwm 0
        for (int i=0 ; i<3 ; i++)
        {
            for (int fan=0 ; fan<4 ; fan++)
            {
                int measureValue = measureRPM(possibleTachoPins[fan]);
                minRPMs[fan] = (minRPMs[fan] < measureValue) ? measureValue : minRPMs[fan];
            }
        }
        
        for (int fan=0 ; fan<4 ; fan++)
        {
            //Serial.print("speed up fan on pin "); Serial.println(pwmPins[fan]);
            setPWM(pwmPins[fan], 100);
            delay(4000);

            //Serial.println("search for max differ rpm.");
            int maxDiff = 0;
            int key = 0;
            int uniqueDiff = true;
            for (int tacho=0 ; tacho<4 ; tacho++)
            {
                int rpm = measureRPM(possibleTachoPins[tacho]);
                int diff = rpm - minRPMs[tacho];
                //Serial.print("old minRPMs[tacho] was: ");  Serial.println(minRPMs[tacho]);
                //Serial.print("slot "); Serial.print(tacho); Serial.print(" is pin ");
                //Serial.print(possibleTachoPins[tacho]); Serial.print(" has "); Serial.print(rpm);
                //Serial.print(" RPM");  Serial.print(" differs "); Serial.println(diff);

                if (diff <= diffThreshold) continue; // skip non differs
              
                if (diff > maxDiff)
                {
                    maxDiff = diff;
                    key = tacho;
                }
                else if (diff == maxDiff) uniqueDiff = false;
            }

            if (maxDiff > 0 && uniqueDiff)
            {
                bool unique = true;
                for(int i=0; i<4 ; i++) if (tachoPins[i]==possibleTachoPins[key]) unique = false;
                if(unique) 
                {
                    tachoPins[fan] = possibleTachoPins[key]; // save pin
                    maxRPMs[fan] = minRPMs[fan] + maxDiff;
                    Serial.print("found: fan"); Serial.print(fan); Serial.print(" pwm_pin="); Serial.print(pwmPins[fan]);
                    Serial.print(" tacho_pin="); Serial.print(tachoPins[fan]); Serial.print(" has min_rpm=");
                    Serial.print(minRPMs[fan]); Serial.print(" RPM, at 100 pwm it has ");
                    Serial.print(maxDiff); Serial.println(" more RPM.");
                }
            }
            setPWM(pwmPins[fan], 0);
        }

        if (foudAtleastOneFan()) return;
        Serial.println("no fans found. need at least one fan. re-checking in 5s.");
        delay(5000);
    }
}

void fanCheck()
{
    Serial.println("starting fan check, may take a bit...");

    for (int fan=0 ; fan<4 ; fan++)
    {
        if (tachoPins[fan] < 0) continue; // skip not existing fans
        //Serial.print("slowing down fan "); Serial.print(fan); Serial.print(" at pin "); Serial.println(pwmPins[fan]);
        setPWM(pwmPins[fan], 0);
        delay(2000);
        
        int minRPM = minRPMs[fan];
        int minPWM = 0;
        //Serial.println("speeding up the fan, step by step.");

        for (int pwm = 0; pwm <= 100; pwm = pwm + 5)
        {
            //Serial.print("set at fan index "); Serial.print(fan);
            //Serial.print(" which has tach on pin"); Serial.print(tachoPins[fan]);
            //Serial.print(" to pwm value "); Serial.println(pwm);
            setPWM(pwmPins[fan], pwm);
            delay(1000);
            //int rpm = measureRPM(tachoPins[fan], 100); // the problem exists, if the rpm is too low at the first measure
            int rpm = measureRPM(tachoPins[fan], 100);
            //Serial.print("after 1 sec i measure "); Serial.print(rpm); Serial.println(" RPM");
      
            if (rpm == minRPM || rpm < minRPM) // this fixes the case than one measure get lower rpm than minimal rpm.
            {
                // found a higher pwm for same rpm
                minPWM = pwm;
            }
            else break; // rpm raised, last rpm ist min pwm for lowest rpm>0, skip here
        }

        setPWM(pwmPins[fan], 0);
        minPWMs[fan] = minPWM; // save value
        Serial.print("Mininal RPM for fan"); Serial.print(fan);
        Serial.print(" is "); Serial.print(minRPM);
        Serial.print(" at PWM value "); Serial.println(minPWM);
    }
}

void configureTimer()
{
    // Configure Timer 1 for PWM @ 25 kHz.
    TCCR1A = 0;   // undo the configuration done by...
    TCCR1B = 0;   // ...the Arduino core library...
    TCCR1C = 0;   // ...for pin 9, 10 and 11.
    TCNT1  = 0;   // reset timer
    // non-inverted PWM on ch. A, B and C, mode 10: ph. correct PWM, TOP = ICR1
    TCCR1A = _BV(COM1A1) | _BV(COM1B1) | _BV(COM1C1) | _BV(WGM11);
    TCCR1B = _BV(WGM13) | _BV(CS10); // prescaler = 1
    TCCR1C = _BV(WGM13) | _BV(CS10); // on PIN 9,10,11
    ICR1   = 320;       // TOP = 320, @16Mhz CPU -> 25kHz PWM
    // Configure Timer 3
    TCCR3A = 0;
    TCNT3  = 0;
    TCCR3A = _BV(COM1A1) | _BV(WGM11);
    TCCR3B = _BV(WGM13) | _BV(CS10);
    ICR3   = 320;
}

void showConfig(bool compact = false)
{
    if (compact)
    {
        // nagios perf data format: 'label'=value[UOM];[warn];[crit];[min];[max] 'label'=value[UOM];[warn];[crit];[min];[max]
        String pwm = "", rpm = "";
        for (int fan=0 ; fan<4 ; fan++)
        {
            pwm = pwm + "pwm" + fan + "=" + ((pwmValues[fan]>0)?pwmValues[fan]:0) + ";;;0;100 ";
            rpm = rpm + "rpm" + fan + "=" + measureRPM(tachoPins[fan]) + ";;" + minRPMs[fan] + ";" + maxRPMs[fan] + "; ";
        }
        String tmp = "";
        for (int i=0 ; i<5 ; i++)
        {
            float tempC = sensors.getTempCByIndex(i);
            tmp = tmp + "tmp" + i + "=" + ((tempC>0)?tempC:0) + ";;;;; ";
        }
        Serial.println(pwm + rpm + tmp);
    }
    else
    {
        Serial.println("--- current config ---");
        for (int fan=0 ; fan<4 ; fan++)
        {
            if (tachoPins[fan] == -1)
            {
                Serial.print("no tacho found for fan "); Serial.print(fan);
                Serial.print(" at pin  "); Serial.println(pwmPins[fan]);
            }
            else
            {
                Serial.print("tacho for fan "); Serial.print(fan);
                Serial.print(" at pin "); Serial.print(pwmPins[fan]);
                Serial.print(" was found on pin "); Serial.print(tachoPins[fan]);
                Serial.print(" the minimal (with rpm>0) pwm value "); Serial.println(minPWMs[fan]);
            }
        }
        Serial.println("--- current config end ---");
    }
}

void setup()
{
    delay(2000);
    Serial.begin(9600);
    Serial.println("fan control setup start.");

    Serial.println("set input pins...");
    for (int i=0 ; i<4 ; i++) pinMode(possibleTachoPins[i], INPUT);

    Serial.println("set output pins...");
    for (int i=0 ; i<4 ; i++) pinMode(pwmPins[i], OUTPUT);

    Serial.println("configure timer for 25kHz PWM...");
    configureTimer();

    // set everything to zero
    for (int fan=0 ; fan<4 ; fan++) setPWM(pwmPins[fan],0);
    //setPWM( 5,25); setPWM( 9,50); setPWM(10,75); setPWM(11,85);

    setupTemperatureSensors();
    assignPinToFan();
    fanCheck();
    showConfig();

    // set fans to minmal pwms (every temp change will alter then)
    for (int fan=0 ; fan<4 ; fan++) setPWM(pwmPins[fan], minPWMs[fan]);

    Serial.println("fan control setup done.");
}

void loop()
{
    if ( millis() > (lastMeasure+(MeasureIntervalInSeconds*1000)) )
    {
        // new measure interval started
        // Serial.print("measure interval started at "); Serial.println(millis());
        float curTemp = getHighestTemp();

        if (curTemp != lastTemp)
        {
            //Serial.print("temp changed from "); Serial.print(lastTemp); Serial.print(" to "); Serial.println(curTemp);

            for (int fan=0 ; fan<4 ; fan++)
            {

                if (tachoPins[fan] < 0) continue; // skip not existing fans
                //Serial.print("for fan index "); Serial.println(fan);

                int newPWM = -1;
                if (curTemp<tempTheFansStartsToIncPWM)
                {
                    newPWM = minPWMs[fan];
                }
                else if (curTemp>=tempTheFanShouldSetToMaxPWM)
                {
                    newPWM = 100;
                }
                else
                {
                    // calc new pwm value
                    //Serial.println("pwm has to be calculated");
                    int steps = ((tempTheFanShouldSetToMaxPWM - tempTheFansStartsToIncPWM) / temperaturePrecisionInDegree) + 1;
                    //Serial.print("steps "); Serial.println(steps);
                    int pwmRange = 100 -  minPWMs[fan];
                    //Serial.print("pwm range "); Serial.println(pwmRange);
                    int pwmIncPerStep = pwmRange / steps;
                    //Serial.print("pwmIncPerStep "); Serial.println(pwmIncPerStep);
                    int stepsToGo = ((curTemp - tempTheFansStartsToIncPWM) / temperaturePrecisionInDegree) + 1;
                    //Serial.print("stepsToGo "); Serial.println(stepsToGo);
                    newPWM = minPWMs[fan] + (stepsToGo * pwmIncPerStep);
                }

                // set the pwm if it changed
                //Serial.print("new pwm value should be "); Serial.println(newPWM);
                if (newPWM != pwmValues[fan])
                {
                    //Serial.print("            !!!!! Differs, so will be set to "); Serial.println(newPWM);
                    pwmValues[fan] = newPWM;
                    setPWM(pwmPins[fan],newPWM);
                }
            }
            lastTemp = curTemp;
        }
        lastMeasure = millis();
    }

    if (Serial.available() > 0)
    {
        // flush input buffer
        while(Serial.available() > 0)
        {
            char t = Serial.read();
            if (t=='r') softReset(100);
        }
        showConfig(true);
    }
}